from .state import State
from ..utils.exceptions import ParseResultsError


class ParseResultsState(State):
    """
    The state which is used to parse the results
    """

    def __init__(self, results):
        self.result = self.__parse(results)

    def __parse(self, results):
        if isinstance(results, bytes):
            results = results.decode("utf-8").strip()
        return results

    def on_event(self):
        return self.result


from pypsrp.client import Client
from pypsrp.powershell import PowerShell, RunspacePool
from pypsrp.wsman import WSMan
from pypsrp.shell import WinRS 

from .state import State
from .parseresults import ParseResultsState


class InnvocationState(State):
    """
    The state which indicates the invocation of a command
    """

    __win_client = None
    __win_error_list = []

    def __handle_windows_errors(self, stream):
        if stream.error:
            for item in stream.error:
                return str(item)
                
        if stream.debug:
            for item in stream.debug:
                self.__win_error_list.append(item)
        if stream.information:
            for item in stream.information:
                self.__win_error_list.append(item)
        if stream.verbose:
            for item in stream.verbose:
                self.__win_error_list.append(item)
        if stream.warning:
            for item in stream.warning:
                self.__win_error_list.append(item)
        
    def __create_win_client(self, hostinfo):
        self.__win_client = Client(
            hostinfo.hostname,
            username=hostinfo.username,
            password=hostinfo.password,
            ssl=hostinfo.verify_ssl
        )

    def __invoke_cmd(self, command):
        if not self.__win_client:
            self.__create_win_client(self.hostinfo)
        stdout, stderr, rc = self.__win_client.execute_cmd(command)
        # NOTE: rc (return code of process) should equal 0 but we are not adding logic here this is handled int he ParseResultsState class
        if stderr:
            print('{host} responded with the following message(s): {message}'.format(
                host=self.hostinfo.hostname,
                message=stderr
            ))
        return ParseResultsState(stdout)

    def __invoke_powershell(self, command):
        if not self.__win_client:
            self.__create_win_client(self.hostinfo)
        output, streams, had_errors = self.__win_client.execute_ps(command)
        if had_errors:
            print('{host} responded with the following message(s): {message}'.format(
                host=self.hostinfo.hostname,
                message=self.__handle_windows_errors(streams)
            ))
        return ParseResultsState(output)

    def __invoke_ssh(self,command):
        import paramiko
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if hasattr(self.hostinfo, 'ssh_key_path'):
            ssh.connect(
                self.hostinfo.hostname,
                port=self.hostinfo.port,
                username=self.hostinfo.username,
                pkey=self.hostinfo.ssh_key_path
            )
        elif hasattr(self.hostinfo, 'password'):
            ssh.connect(
                self.hostinfo.hostname,
                port=self.hostinfo.port,
                username=self.hostinfo.username,
                password=self.hostinfo.password
            )
        else:
            raise AttributeError('Please provide either a ssh_key_path or a password')
        out = None
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command)
        out = ssh_stdout.read()
        ssh_stdin.flush()
        ssh.close()
        return ParseResultsState(out)

    def invoke(self, hostinfo, command_type, command):
        self.hostinfo = hostinfo
        if command_type == 'powershell':
            result = self.__invoke_powershell(command)
        elif command_type == 'cmd':
            result = self.__invoke_cmd(command)
        elif command_type == 'ssh':
            result = self.__invoke_ssh(command)
        return result


from .states.creation import CreationState


class Rudder(object):
    """ 
    A simple state machine that mimics the functionality of a device from a 
    high level.
    """

    def __init__(self):
        """ Initialize the components. """

        # Start with a default state.
        self.state = CreationState()

    def execute(self, hostinfo, command_type: str, command: str):
        """The execute method is the main entry point to execute the provided command on a remote host.

        You must create a host object and pass it to this method along with the command_type and command to execute.

        Args:
            hostinfo (HostInfo): A host information object must be created and passed into this method
            command_type (str): A command_type of powershell or cmd or ssh is needed
            command (str): The command to run on a specified remote host

        Returns:
            str: Returns the results from a remote command
        """        
        while True:
            if str(self.state) == 'CreationState':
                self.state = self.state.on_event(command_type, command)
            if str(self.state) == 'InnvocationState':
                self.state = self.state.invoke(hostinfo, command_type, command)
            if str(self.state) == 'ParseResultsState':
                return self.state.on_event()



from .state import State
from .innvocation import InnvocationState


class CreationState(State):
    """
    The state which is used to modify commands provided to rudder
    """

    def powershell(self, event):
        command = None
        if event:
            if '\n' in event or '\r' in event:
                if '\n' in event:
                    command = event.replace('\n', '; ')
                if '\r' in event:
                    if command:
                        command = command.replace('\r', '; ')
                    else:
                        command = event.replace('\r', '; ')
            return InnvocationState()

    def cmd(self):
        return InnvocationState()

    def ssh(self):
        return InnvocationState()

    def on_event(self, command_type, command):
        if command_type == 'powershell':
            return self.powershell(command)
        elif command_type == 'cmd':
            return self.cmd()
        elif command_type == 'ssh':
            return self.ssh()
        return self


import sys


class HostInfo(object):

    """A HostInfo object is used when executing remote commands on hosts
    """

    def windows(self, hostname, username, password, verify_ssl=False):
        """The windows method provides information to connect to a remote windows host using the PowerShell Remoting Protocol and WinRM

        Args:
            hostname (str): A single hostname to connect to.  Either DNS or IP address.
            username (str): A username that has access to remote into this machine
            password (str): A password that works with the username provided
            verify_ssl (bool, optional): Whether or not you are using TLS/SSL certificates. Defaults to False.

        Returns:
            HostInfo: Returns a HostInfo object used to connect to a remote Windows host
        """
        self.hostname = hostname
        self.username = username
        self.password = password
        self.verify_ssl = verify_ssl
        return self

    def linux(self, hostname, username, ssh_key_path=None, password=None, port=22):
        """The linux method provides information to execute code on a remote host using SSH.  This can be Linux or macOS.

        Args:
            hostname (str): A single hostname to connect to.  Either DNS or IP address.
            username (str): A username that has access to remote into this machine
            password (str, optional): A password that works with the username provided.
            ssh_key_path (str, optional): A ssh key path to authenticate to a remote host
            port (int, optional): The port to use when SSHing. Defaults to 22.

        Returns:
            HostInfo: Returns a HostInfo object used to connect to a remote Linux or macOS host
        """
        self.hostname = hostname
        self.username = username
        
        if ssh_key_path:
            self.ssh_key_path = ssh_key_path
        elif password:
            self.password = password
        else:
            raise AttributeError('Please provide a ssh key path or a password!')
        self.port = port
        return self



from setuptools import setup, find_packages

def parse_requirements(requirement_file):
    with open(requirement_file) as f:
        return f.read().splitlines()

version = dict()
with open("./rudder/utils/version.py") as fp:
    exec(fp.read(), version)


setup(
    name='rudder',
    version=version['__version__'],
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    description='A Python package to execute code remotely to multiple operating system platforms',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=parse_requirements('requirements.txt'),
    keywords=['rudder', 'winrm', 'ssh', 'cmd', 'powershell'],
    url='https://github.com/MSAdministrator/rudder',
    author='MSAdministrator',
    author_email='rickardja@live.com',
    python_requires='>3.6'
)